/**
Compare Laplacian and Taubin smoothing on 
a regularly sampled 1D domain.

Orion Sky Lawlor, olawlor@acm.org, 2003/9/23
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* Size of domain, in cells. */
#define N 100000

class fn {
	// Store the domain with two "ghost" cells on the ends.
	double v[1+N+1];
public:
	fn(void) {v[0]=v[N+1]=1.0;}
	
	/// Extract values from the interior of the domain 
	inline double &operator[](int i) {return v[1+i];}
	inline double operator[](int i) const {return v[1+i];}
};

double randFloat(void) {
	return (rand()&1023)*(1.0/(double)1023);
}

/**
 Initialize this function with a frequency ramp,
   from 0 to nyquist.
 Logic: Take t going from 0 (left end) to 1.0 (right end)
   Then if freqency linearly increases from 0 to 
   nyquist (N/2.0 Hz), then
        freq = (N/2.0) * t
   Now integrate with respect to t to get revolutions
        phaseR = (N/2.0) * t^2/2
   Finally, convert to radians
        phase = (2*M_PI) * (N/2.0) * t^2/2
*/
void makeRamp(fn &dest) {
	for (int i=0;i<N;i++)
	{
		double t=i*(1.0/N);
		double phase=(2*M_PI)*(N/2.0)*0.5*t*t+0.0*randFloat();
		dest[i]=cos(phase);
	}
}

/**
 Initialize this function with a simple step.
*/
void makeStep(fn &dest) {
	for (int i=-1;i<=N;i++)
	{
		if (i<N/2) dest[i]=1.0;
		else dest[i]=0.0;
	}
}

/**
 Apply Laplacian smoothing, using the umbrella operator
 to approximate the Laplacian:
    dest = src + lambda * umbrella(src);
*/
void smoothLaplace(const fn &src,fn &dest,double lambda) {
	for (int i=0;i<N;i++) {
		double umbrella=0.5*(src[i-1]+src[i+1])-src[i];
		dest[i]=src[i]+lambda*umbrella;
	}
	dest[-1]=src[-1];
	dest[N]=src[N];
}

/**
 Apply Taubin smoothing in-place, with some fixed constants.
*/
void smoothTaubin(fn &inOut) {
	double t_lambda=0.6307;
	double t_mu=-0.6732;
	fn tmp;
	smoothLaplace(inOut,tmp,t_lambda);
	smoothLaplace(tmp,inOut,t_mu);
}

/**
 Write this function to this text file, lines of the form:
     <time, seconds>  <amplitude>
*/
void write(const fn &src,const char *dest) {
	FILE *f=fopen(dest,"w");
	for (int i=0;i<N;i++) {
		double t=i*(1.0/N);
		fprintf(f,"%f %f\n",t,(double)src[i]);
	}
	fclose(f);
	printf("Wrote %d points to %s\n",N,dest);
}

/**
 Write the ratio of these two functions to this text file.
*/
void writeRatio(const fn &src,const fn &ref,const char *dest) {
	fn ratio;
	double vp=0.0;
	for (int i=0;i<N;i++) {
		double v=src[i]/ref[i];
		if (fabs(ref[i])<0.3) v=vp;
		ratio[i]=v;
		vp=v;
	}
	write(ratio,dest);
}


int main() {
	int t;
	int nIter=1000;
// Start with a frequency ramp
	fn ramp; 
	makeRamp(ramp);
	writeRatio(ramp,ramp,"ramp.sig");
	
// Apply laplacian smoothing
	double l_lambda=0.7;
	fn lap=ramp;
	for (t=0;t<nIter;t++) {
		fn tmp;
		smoothLaplace(lap,tmp,l_lambda);
		lap=tmp;
		if (t==0) writeRatio(lap,ramp,"laplace_1.sig");
		if (t==99) writeRatio(lap,ramp,"laplace_100.sig");
	}
	writeRatio(lap,ramp,"laplace_1000.sig");
	
// Apply Taubin smoothing
	fn taubin=ramp;
	for (t=0;t<nIter;t++) {
		smoothTaubin(taubin);
		if (t==0) writeRatio(taubin,ramp,"taubin_1.sig");
		if (t==99) writeRatio(taubin,ramp,"taubin_100.sig");
	}
	writeRatio(taubin,ramp,"taubin_1000.sig");

// Start with a step function, and smooth it:
	makeStep(lap);
	write(lap,"lstep.sig");
	for (t=0;t<1000;t++) smoothLaplace(lap,lap,l_lambda);
	write(lap,"lstep_1000.sig");
	
	makeStep(taubin);
	write(taubin,"tstep.sig");
	for (t=0;t<1000;t++) smoothTaubin(taubin);
	write(taubin,"tstep_1000.sig");
	
	return 0;
}



